﻿//---------------------------------------------------------------------
// <copyright file="CommandLineOptions.cs" company="Microsoft">
//      Copyright (C) Microsoft Corporation. All rights reserved. See License.txt in the project root for license information.
// </copyright>
//---------------------------------------------------------------------

using System.Collections.Generic;
using System.IO;
using CommandLine;
using CommandLine.Text;

namespace ResultsComparer
{
    public class CommandLineOptions
    {
        [Option("type", HelpText = "The type of reports to compare. Possible values are 'bdn' for BenchmarkDotNet, 'vsAllocs' for Visual Studio .NET Object Allocator")]
        public string Comparer { get; set; }

        [Option("base", HelpText = "Path to the folder/file with base results.")]
        public string BasePath { get; set; }

        [Option("diff", HelpText = "Path to the folder/file with diff results.")]
        public string DiffPath { get; set; }

        [Option("threshold", HelpText = "Threshold for Statistical Test. Examples: 5%, 10ms, 100ns, 1s.")]
        public string StatisticalTestThreshold { get; set; }

        [Option("noise", HelpText = "Noise threshold for Statistical Test. The difference for 1.0ns and 1.1ns is 10%, but it's just a noise. Examples: 0.5ns 1ns.", Default = "0.3ns" )]
        public string NoiseThreshold { get; set; }
        
        [Option("metric", HelpText = "The metric to compare from the reports. Optional.")]
        public string Metric { get; set; }

        [Option("top", HelpText = "Filter the diff to top/bottom N results. Optional.")]
        public int? TopCount { get; set; }

        [Option("csv", HelpText = "Path to exported CSV results. Optional.")]
        public FileInfo CsvPath { get; set; }

        [Option("xml", HelpText = "Path to exported XML results. Optional.")]
        public FileInfo XmlPath { get; set; }

        [Option('f', "filter", HelpText = "Filter the benchmarks by name using glob pattern(s). Optional.")]
        public IEnumerable<string> Filters { get; set; }

        [Option("full-id", HelpText = "Display the full benchmark name id. Optional.")]
        public bool FullId { get; set; }

        [Usage(ApplicationAlias = "")]
        public static IEnumerable<Example> Examples
        {
            get
            {
                yield return new Example(@"Compare the results stored in 'C:\results\ODataWriterBenchmarksBefore.json' (base) vs 'C:\results\ODataWriterBenchmarksAfter.json' (diff) using 5% threshold.",
                    new CommandLineOptions { Comparer = "bdn", BasePath = @"C:\results\ODataWriterBenchmarksBefore.json", DiffPath = @"C:\results\ODataWriterBenchmarksAfter.json", StatisticalTestThreshold = "5%" });
                yield return new Example(@"Compare the results stored in 'C:\results\win' (base) vs 'C:\results\unix' (diff) using 5% threshold and show only top/bottom 10 results.",
                    new CommandLineOptions { Comparer = "bdn", BasePath = @"C:\results\win", DiffPath = @"C:\results\unix", StatisticalTestThreshold = "5%", TopCount = 10 });
                yield return new Example(@"Compare the results stored in 'C:\results\win' (base) vs 'C:\results\unix' (diff) using 5% threshold and 0.5ns noise filter.",
                    new CommandLineOptions { Comparer = "bdn", BasePath = @"C:\results\win", DiffPath = @"C:\results\unix", StatisticalTestThreshold = "5%", NoiseThreshold = "0.5ns" });
                yield return new Example(@"Compare the System.Math benchmark results stored in 'C:\results\ubuntu16' (base) vs 'C:\results\ubuntu18' (diff) using 5% threshold.",
                    new CommandLineOptions { Comparer = "bdn", Filters = new[] { "System.Math*" }, BasePath = @"C:\results\win", DiffPath = @"C:\results\unix", StatisticalTestThreshold = "5%" });
            }
        }
    }
}